"""Class to deal with scenes?"""

from copy import deepcopy
from typing import Optional, Tuple

import numpy as np
import numpy.typing as npt
from scipy import sparse
from tqdm import tqdm

from .psf import PSF
from .sparsewarp import ROISparseWarp3D, SparseWarp3D
from .utils import downsample as downsample_array
from .utils import prep_for_add

__all__ = ["ROIScene", "Scene", "TraceScene"]


class Scene(object):
    def __init__(
        self,
        locations: npt.ArrayLike,
        psf: PSF = None,
        shape: Tuple = (100, 100),
        corner: Tuple = (0, 0),
        scale: int = 1,
    ):
        if locations.shape[1] != 2:
            raise ValueError("`locations` must have shape (n, 2).")
        self.locations = locations
        #        self.shape = shape
        #        self.corner = corner
        if psf is None:
            psf = PSF.from_name("visda", scale=scale)
        self.psf = psf
        self.scale = self.psf.scale
        self.shape = tuple(np.asarray(shape) * self.scale)
        self.corner = tuple(np.asarray(corner) * self.scale)
        self.rb, self.cb, self.prf = [], [], []
        self.ntargets = len(self.locations)
        self._get_X()

    def _get_ar(self, flux, delta_pos=None, quiet=True):
        nt = flux.shape[1]
        ar = np.zeros((nt, *self.shape))
        if delta_pos is not None:
            jitterdec = (delta_pos - 0.5) % 1 - 0.5
            jitterint = np.round(delta_pos - jitterdec).astype(int)  # + 1
            unique_coordinates, unique_indices = np.unique(
                jitterint.T, axis=0, return_inverse=True
            )
            for index, coord in tqdm(
                enumerate(unique_coordinates),
                total=len(unique_coordinates),
                leave=True,
                position=0,
                desc="Modeling Pixel Positions",
                disable=quiet,
            ):
                self.X.translate(tuple(coord))
                self.dX0.translate(tuple(coord))
                self.dX1.translate(tuple(coord))
                tdxs = np.where(unique_indices == index)[0]
                ar[tdxs] = self.X.dot(flux[:, tdxs])
                ar[tdxs] += self.dX0.dot(
                    flux[:, tdxs] * -jitterdec[0, tdxs]
                ) + self.dX1.dot(flux[:, tdxs] * -jitterdec[1, tdxs])
        else:
            ar = self.X.dot(flux)
        return ar

    def __repr__(self):
        return f"Scene Object [{self.psf.__repr__()}] Detector Size: {self.shape}, ntargets: {self.ntargets}"

    def __len__(self):
        return len(self.locations)

    def _get_X(self):
        row, col, data, grad0, grad1 = [], [], [], [], []
        for location in self.locations * self.scale:
            r, c, ar, g0, g1 = self.psf.prf(
                row=location[0], column=location[1], gradients=True
            )
            row.append(r[:, None] * np.ones(c.shape[0], int))
            col.append(c[None, :] * np.ones(r.shape[0], int)[:, None])
            grad0.append(g0)
            grad1.append(g1)
            data.append(ar)
        data, row, col, grad0, grad1 = (
            np.asarray(data),
            np.asarray(row),
            np.asarray(col),
            np.asarray(grad0),
            np.asarray(grad1),
        )
        self.X = SparseWarp3D(
            data.transpose([1, 2, 0]),
            row.transpose([1, 2, 0]) - self.corner[0],
            col.transpose([1, 2, 0]) - self.corner[1],
            imshape=self.shape,
        )
        self.dX0 = SparseWarp3D(
            grad0.transpose([1, 2, 0]),
            row.transpose([1, 2, 0]) - self.corner[0],
            col.transpose([1, 2, 0]) - self.corner[1],
            imshape=self.shape,
        )
        self.dX1 = SparseWarp3D(
            grad1.transpose([1, 2, 0]),
            row.transpose([1, 2, 0]) - self.corner[0],
            col.transpose([1, 2, 0]) - self.corner[1],
            imshape=self.shape,
        )
        return

    def model(
        self,
        flux: npt.ArrayLike,
        delta_pos: Optional[npt.ArrayLike] = None,
        quiet: bool = False,
        downsample: bool = True,
    ) -> npt.ArrayLike:
        """
        Parameters:
        -----------
        flux : npt.ArrayLike
            Array of flux values with shape (ntargets, ntimes)
        delta_pos : npt.ArrayLike
            Array of jitter values in row and column, has shape (2, ntimes)
        """
        if flux.ndim == 1:
            flux = flux[:, None]
        if flux.shape[0] != self.ntargets:
            raise ValueError("`flux` must be an array with shape (ntargets x ntimes).")
        # nt = flux.shape[1]
        if delta_pos is not None:
            delta_pos = deepcopy(delta_pos) * float(self.scale)
        #     if delta_pos.ndim == 1:
        #         delta_pos = delta_pos[:, None]
        #     if (delta_pos.shape[0] != 2) | (delta_pos.shape[1] != nt):
        #         raise ValueError(
        #             "`delta_pos` must be an array with shape (2 x ntimes)."
        #         )

        # if delta_pos is not None:
        #     ar = np.zeros((nt, *self.shape))
        #     jitterdec = (delta_pos - 0.5) % 1 - 0.5
        #     jitterint = np.round(delta_pos - jitterdec).astype(int)  # + 1
        #     unique_coordinates, unique_indices = np.unique(
        #         jitterint.T, axis=0, return_inverse=True
        #     )
        #     for index, coord in enumerate(unique_coordinates):
        #         self.X.translate(tuple(coord))
        #         self.dX0.translate(tuple(coord))
        #         self.dX1.translate(tuple(coord))
        #         grad_ar = deepcopy(self.dX0)
        #         tdxs = np.where(unique_indices == index)[0]
        #         for tdx in tqdm(
        #             tdxs,
        #             desc=f"Time index {index + 1}/{len(unique_coordinates)}",
        #             leave=True,
        #             position=0,
        #             disable=quiet,
        #         ):
        #             grad_ar.data = (
        #                 deepcopy(self.dX0.data) * -jitterdec[0, tdx]
        #                 + deepcopy(self.dX1.data) * -jitterdec[1, tdx]
        #             )
        #             ar[tdx] += self.X.dot(flux[:, tdx].ravel())[0]
        #             ar[tdx] += grad_ar.dot(flux[:, tdx].ravel())[0]
        # else:
        #     ar = self.X.dot(flux)

        ar = self._get_ar(flux=flux, delta_pos=delta_pos, quiet=quiet)

        if self.scale == 1:
            return ar
        if downsample:
            return downsample_array(ar, self.scale)
        else:
            return ar

    def fit_images(self, imgs, prior_mu=None, prior_sigma=None, fit_shifts=False):
        """Fit a stack of images with the PRF model"""
        if imgs.ndim == 2:
            imgs = imgs[None, :, :]
        if imgs.shape[1:] != self.shape:
            raise ValueError(f"Must supply an image with shape {self.shape}.")

        sparse_imgs = []
        for img in imgs:
            data = img[
                self.X.subrow.ravel() % self.shape[0],
                self.X.subcol.ravel() % self.shape[1],
            ].reshape(self.X.subrow.shape)
            bad = (
                (self.X.subrow < 0)
                | (self.X.subcol < 0)
                | (self.X.subrow > self.shape[0])
                | (self.X.subrow > self.shape[1])
            )
            data[bad] = 0
            d = SparseWarp3D(data, self.X.subrow, self.X.subcol, self.X.imshape)
            d = d.tocsr()
            sparse_imgs.append(d.max(axis=1))
        y = sparse.hstack(sparse_imgs)

        A = self.X.tocsr()
        if prior_mu is not None:
            pm = prior_mu.copy()
        else:
            pm = np.zeros(A.shape[1])
        if prior_sigma is not None:
            ps = prior_sigma.copy()
        else:
            ps = np.ones(A.shape[1]) * np.inf
        if fit_shifts:
            if prior_mu is None:
                raise ValueError(
                    "You can only fit shifts if a flux estimate is provided via `prior_mu`."
                )
            A = sparse.hstack(
                [
                    A,
                    -self.dX0.tocsr().dot(sparse.csr_matrix(prior_mu).T),
                    -self.dX1.tocsr().dot(sparse.csr_matrix(prior_mu).T),
                ]
            )
            pm = np.hstack([pm, 0, 0])
            ps = np.hstack([ps, np.inf, np.inf])
        sigma_w_inv = A.T.dot(A).toarray()

        B = A.T.dot(y).toarray()
        w = np.linalg.solve(
            sigma_w_inv + np.diag(1 / ps**2), B + pm[:, None] / ps[:, None] ** 2
        )
        werr = np.linalg.inv(sigma_w_inv).diagonal() ** 0.5
        if fit_shifts:
            return w[:-2], werr[:-2], w[-2:], werr[-2:]
        return w, werr, np.asarray([0, 0]), np.asarray([0, 0])


class ROIScene(Scene):
    def __init__(
        self,
        locations: npt.ArrayLike,
        psf: PSF = None,
        shape: Tuple = (100, 100),
        corner: Tuple = (0, 0),
        scale: int = 1,
        nROIs=1,
        ROI_size=(10, 10),
        ROI_corners=[(0, 0)],
    ):
        self.nROIs = nROIs
        self.ROI_size = ROI_size
        self.ROI_corners = ROI_corners
        super().__init__(
            locations=locations, psf=psf, shape=shape, corner=corner, scale=scale
        )

    def _get_X(self):
        row, col, data, grad0, grad1 = [], [], [], [], []
        for location in self.locations * self.scale:
            r, c, ar, g0, g1 = self.psf.prf(
                row=location[0], column=location[1], gradients=True
            )
            row.append(r[:, None] * np.ones(c.shape[0], int))
            col.append(c[None, :] * np.ones(r.shape[0], int)[:, None])
            grad0.append(g0)
            grad1.append(g1)
            data.append(ar)
        data, row, col, grad0, grad1 = (
            np.asarray(data),
            np.asarray(row),
            np.asarray(col),
            np.asarray(grad0),
            np.asarray(grad1),
        )
        self.X = ROISparseWarp3D(
            data.transpose([1, 2, 0]),
            row.transpose([1, 2, 0]) - self.corner[0],
            col.transpose([1, 2, 0]) - self.corner[1],
            imshape=self.shape,
            nROIs=self.nROIs,
            ROI_size=self.ROI_size,
            ROI_corners=self.ROI_corners,
        )
        self.dX0 = ROISparseWarp3D(
            grad0.transpose([1, 2, 0]),
            row.transpose([1, 2, 0]) - self.corner[0],
            col.transpose([1, 2, 0]) - self.corner[1],
            imshape=self.shape,
            nROIs=self.nROIs,
            ROI_size=self.ROI_size,
            ROI_corners=self.ROI_corners,
        )
        self.dX1 = ROISparseWarp3D(
            grad1.transpose([1, 2, 0]),
            row.transpose([1, 2, 0]) - self.corner[0],
            col.transpose([1, 2, 0]) - self.corner[1],
            imshape=self.shape,
            nROIs=self.nROIs,
            ROI_size=self.ROI_size,
            ROI_corners=self.ROI_corners,
        )
        return

    def _get_ar(self, flux, delta_pos=None, quiet=True):
        nt = flux.shape[1]
        ar = np.zeros((self.nROIs, nt, *self.ROI_size))
        if delta_pos is not None:
            jitterdec = (delta_pos - 0.5) % 1 - 0.5
            jitterint = np.round(delta_pos - jitterdec).astype(int)  # + 1
            unique_coordinates, unique_indices = np.unique(
                jitterint.T, axis=0, return_inverse=True
            )
            for index, coord in tqdm(
                enumerate(unique_coordinates),
                total=len(unique_coordinates),
                leave=True,
                position=0,
                desc="Modeling Pixel Positions",
                disable=quiet,
            ):
                self.X.translate(tuple(coord))
                self.dX0.translate(tuple(coord))
                self.dX1.translate(tuple(coord))
                tdxs = np.where(unique_indices == index)[0]
                ar[:, tdxs] = self.X.dot(flux[:, tdxs])
                ar[:, tdxs] += self.dX0.dot(
                    flux[:, tdxs] * -jitterdec[0, tdxs]
                ) + self.dX1.dot(flux[:, tdxs] * -jitterdec[1, tdxs])
        else:
            ar = self.X.dot(flux)
        return ar

    def fit_images(self, imgs, prior_mu=None, prior_sigma=None, fit_shifts=False):
        """Fit a stack of images with the PRF model"""
        if imgs.ndim == 3:
            imgs = imgs[:, None, :, :]
        if imgs.ndim != 4:
            raise ValueError("Must supply 4D data.")
        if imgs.shape[2:] != self.ROI_size:
            raise ValueError(
                f"Must supply a images with shape (nROIs, ntimes, *ROI_size), at least {(self.nROIs, 1, *self.ROI_size)}."
            )
        if imgs.shape[0] != self.nROIs:
            raise ValueError(
                f"Must supply a images with shape (nROIs, ntimes, *ROI_size), at least {(self.nROIs, 1, *self.ROI_size)}."
            )
        sparse_imgs = []
        R, C = np.meshgrid(
            np.arange(0, self.ROI_size[0]),
            np.arange(0, self.ROI_size[1]),
            indexing="ij",
        )
        row = np.asarray([R + corner[0] for corner in self.ROI_corners]).transpose(
            [1, 2, 0]
        )
        column = np.asarray([C + corner[1] for corner in self.ROI_corners]).transpose(
            [1, 2, 0]
        )

        for img in imgs.transpose([1, 0, 2, 3]):
            data = img.transpose([1, 2, 0])
            d = SparseWarp3D(data=data, row=row, col=column, imshape=self.shape)
            d = d.tocsr()
            sparse_imgs.append(d.max(axis=1))
        y = sparse.hstack(sparse_imgs)

        A = self.X.tocsr()
        if prior_mu is not None:
            pm = prior_mu.copy()
        else:
            pm = np.zeros(A.shape[1])
        if prior_sigma is not None:
            ps = prior_sigma.copy()
        else:
            ps = np.ones(A.shape[1]) * np.inf
        if fit_shifts:
            if prior_mu is None:
                raise ValueError(
                    "You can only fit shifts if a flux estimate is provided via `prior_mu`."
                )
            A = sparse.hstack(
                [
                    A,
                    -self.dX0.tocsr().dot(sparse.csr_matrix(prior_mu).T),
                    -self.dX1.tocsr().dot(sparse.csr_matrix(prior_mu).T),
                ]
            )
            pm = np.hstack([pm, 0, 0])
            ps = np.hstack([ps, np.inf, np.inf])
        sigma_w_inv = A.T.dot(A).toarray()

        B = A.T.dot(y).toarray()
        w = np.linalg.solve(
            sigma_w_inv + np.diag(1 / ps**2), B + pm[:, None] / ps[:, None] ** 2
        )
        werr = np.linalg.inv(sigma_w_inv).diagonal() ** 0.5
        if fit_shifts:
            return w[:-2], werr[:-2], w[-2:], werr[-2:]
        return w, werr, np.asarray([0, 0]), np.asarray([0, 0])

    def __repr__(self):
        return f"ROIScene Object [{self.psf.__repr__()}] Detector Size: {self.shape}, ntargets: {self.ntargets}, nROIs: {self.nROIs}"


class TraceScene(Scene):
    def __init__(
        self,
        locations: npt.ArrayLike,
        psf: PSF = None,
        shape: Tuple = (400, 80),
        corner: Tuple = (0, 0),
        scale: int = 1,
        wav_bin: int = 4,
    ):
        if locations.shape[1] != 2:
            raise ValueError("`locations` must have shape (n, 2).")
        self.locations = locations
        if psf is None:
            psf = PSF.from_name("nirda", scale=scale)
        self.psf = psf
        self.scale = self.psf.scale
        self.shape = tuple(np.asarray(shape) * self.scale)
        self.corner = tuple(np.asarray(corner) * self.scale)
        if "wavelength" not in self.psf.dimension_names:
            raise ValueError(
                "Can only create a trace scene if `PSF` has a `'wavelength'` dimension"
            )
        if not hasattr(self.psf, "trace_pixel"):
            raise ValueError("No trace parameters, you need to set them.")
        self.ntargets = len(self.locations)
        self.rb, self.cb, self.prf = [], [], []
        self.wav_bin = wav_bin
        self._get_Xs(nbin=self.wav_bin)

    def __repr__(self):
        return f"TraceScene Object [{self.psf.__repr__()}]"

    def _get_Xs(self, nbin=4):
        def get_chunked_data(locations, pixels, waves):
            """Loops through pixel trave groups and gets each PRF element into a dictionary"""
            res = {}
            for pdx in range(len(pixels)):
                for location in locations:
                    key = f"({location[0]}, {location[1]}) {pdx}"
                    res[key] = {
                        pixels[pdx][idx].value: self.psf.prf(
                            row=location[0] + pixels[pdx][idx].value,
                            column=location[1],
                            wavelength=waves[pdx][idx],
                            gradients=True,
                        )
                        for idx in range(len(pixels[pdx]))
                    }
                    r = np.asarray(
                        [
                            [res[key][pix][0].min(), res[key][pix][0].max()]
                            for pix in pixels[pdx].value
                        ]
                    )
                    r = (np.min(r[:, 0]), np.max(r[:, 1]))
                    c = np.asarray(
                        [
                            [res[key][pix][1].min(), res[key][pix][1].max()]
                            for pix in pixels[pdx].value
                        ]
                    )
                    c = (np.min(c[:, 0]), np.max(c[:, 1]))
                    res[key]["r"] = r
                    res[key]["c"] = c
            return res

        def collapse(dic, shape):
            """For each location and pixel group, collapses the data into a single, small image"""
            rshape, cshape = shape
            corner = (dic["r"][0], dic["c"][0])
            ar, dar0, dar1 = np.zeros((3, rshape, cshape))
            for key, item in dic.items():
                if key in ["r", "c"]:
                    continue
                rb, cb, arb = prep_for_add(
                    row=item[0],
                    column=item[1],
                    prf=item[2],
                    shape=(rshape, cshape),
                    corner=corner,
                )
                _, _, dar0b = prep_for_add(
                    row=item[0],
                    column=item[1],
                    prf=item[3],
                    shape=(rshape, cshape),
                    corner=corner,
                )
                _, _, dar1b = prep_for_add(
                    row=item[0],
                    column=item[1],
                    prf=item[4],
                    shape=(rshape, cshape),
                    corner=corner,
                )
                ar[rb, cb] += arb
                dar0[rb, cb] += dar0b
                dar1[rb, cb] += dar1b

            R, C = np.mgrid[:rshape, :cshape]
            R += corner[0]
            C += corner[1]
            corr = np.sum(ar)

            return (
                R,
                C,
                ar / corr,
                (dar0 - dar0.mean()) / corr,
                (dar1 - dar1.mean()) / corr,
            )

        pixs, wavs = self.psf.trace_pixel * self.psf.scale, self.psf.trace_wavelength
        dwavs = np.gradient(wavs)
        pixs, wavs, dwavs = (
            np.array_split(pixs, len(pixs) / nbin),
            np.array_split(wavs, len(wavs) / nbin),
            np.array_split(dwavs, len(dwavs) / nbin),
        )

        bounds0 = np.asarray(
            [wav[0].value - dwav[0].value / 2 for wav, dwav in zip(wavs, dwavs)]
        )
        bounds1 = np.hstack(
            [*bounds0[1:], wavs[-1][-1].value + dwavs[-1][-1].value / 2]
        )
        self.bounds = np.asarray([bounds0, bounds1])
        self.nwav = self.bounds.shape[1]
        data, grad0, grad1 = np.zeros((3, len(pixs), len(self.locations), *self.shape))

        res = get_chunked_data(self.locations * self.psf.scale, pixs, wavs)

        r = np.asarray([item["r"] for _, item in res.items()])
        c = np.asarray([item["c"] for _, item in res.items()])
        shape = np.max(r[:, 1] - r[:, 0]), np.max(c[:, 1] - c[:, 0])

        R, C, data, grad0, grad1 = np.zeros((5, len(res), *shape))
        for idx, key in enumerate(res.keys()):
            R[idx], C[idx], data[idx], grad0[idx], grad1[idx] = collapse(
                res[key], shape
            )
        data, grad0, grad1 = data, grad0, grad1
        self.X = SparseWarp3D(
            data.transpose([1, 2, 0]),
            R.transpose([1, 2, 0]) - self.corner[0],
            C.transpose([1, 2, 0]) - self.corner[1],
            imshape=self.shape,
        )

        self.dX0 = SparseWarp3D(
            grad0.transpose([1, 2, 0]),
            R.transpose([1, 2, 0]) - self.corner[0],
            C.transpose([1, 2, 0]) - self.corner[1],
            imshape=self.shape,
        )

        self.dX1 = SparseWarp3D(
            grad1.transpose([1, 2, 0]),
            R.transpose([1, 2, 0]) - self.corner[0],
            C.transpose([1, 2, 0]) - self.corner[1],
            imshape=self.shape,
        )
        self.wavelength = wavs

    # def _get_Xs(self, nbin=4):
    #     pixs, wavs = self.psf.trace_pixel * self.psf.scale, self.psf.trace_wavelength
    #     mask = self.psf.trace_sensitivity/self.psf.trace_sensitivity.max() > 0.0001
    #     pixs, wavs = pixs[mask], wavs[mask]
    #     dwavs = np.gradient(wavs)
    #     pixs, wavs, dwavs = np.array_split(pixs, len(pixs)/nbin),
    #     np.array_split(wavs, len(wavs)/nbin), np.array_split(dwavs, len(dwavs)/nbin)

    #     bounds0 = np.asarray([wav[0].value - dwav[0].value/2 for wav, dwav in zip(wavs, dwavs)])
    #     bounds1 = np.hstack([*bounds0[1:], wavs[-1][-1].value + dwavs[-1][-1].value/2])
    #     self.bounds = np.asarray([bounds0, bounds1])
    #     self.nwav = self.bounds.shape[1]
    #     data, grad0, grad1 = np.zeros((3, len(pixs), len(self.locations), *self.shape))
    #     for pdx, pix, wav in tqdm(zip(range(len(pixs)), pixs, wavs), total=len(pixs)):
    #         for ldx, location in enumerate(self.locations * self.psf.scale):
    #             for idx in range(len(pix)):
    #                 r, c, ar, g0, g1 = self.psf.prf(
    #                     row=location[0] + pix.value[idx],
    #                     column=location[1],
    #                     wavelength=wav[idx],
    #                     gradients=True,
    #                 )
    #                 row, col, ar = prep_for_add(r, c, ar, shape=self.shape, corner=self.corner)
    #                 _, _, g0 = prep_for_add(r, c, g0, shape=self.shape, corner=self.corner)
    #                 _, _, g1 = prep_for_add(r, c, g1, shape=self.shape, corner=self.corner)
    #                 data[pdx, ldx, row, col] += ar
    #                 grad0[pdx, ldx, row, col] += g0
    #                 grad1[pdx, ldx, row, col] += g1
    #     data, grad0, grad1 = np.vstack(data), np.vstack(grad0), np.vstack(grad1)
    #     R, C = np.meshgrid(np.arange(self.shape[0]), np.arange(self.shape[1]), indexing='ij')
    #     R, C = R[None, :, :] * np.ones(data.shape) - self.corner[0], C[None, :, :] * np.ones(data.shape) - self.corner[1]

    #     self.X = SparseWarp3D(
    #                 data.transpose([1, 2, 0]),
    #                 R.transpose([1, 2, 0]),
    #                 C.transpose([1, 2, 0]),
    #                 imshape=self.shape,
    #     )
    #     self.dX0 = SparseWarp3D(
    #                 grad0.transpose([1, 2, 0]),
    #                 R.transpose([1, 2, 0]),
    #                 C.transpose([1, 2, 0]),
    #         imshape=self.shape,
    #     )
    #     self.dX1 = SparseWarp3D(
    #                 grad1.transpose([1, 2, 0]),
    #                 R.transpose([1, 2, 0]),
    #                 C.transpose([1, 2, 0]),
    #         imshape=self.shape,
    #     )
    #     return
    # pixs, wavs = self.psf.trace_pixel * self.psf.scale, self.psf.trace_wavelength
    # mask = self.psf.trace_sensitivity/self.psf.trace_sensitivity.max() > 0.0001
    # pixs, wavs = pixs[mask], wavs[mask]
    # dwavs = np.gradient(wavs)
    # pixs, wavs, dwavs = np.array_split(pixs, len(pixs)/nbin), np.array_split(wavs, len(wavs)/nbin), np.array_split(dwavs, len(dwavs)/nbin)

    # bounds0 = np.asarray([wav[0].value - dwav[0].value/2 for wav, dwav in zip(wavs, dwavs)])
    # bounds1 = np.hstack([*bounds0[1:], wavs[-1][-1].value + dwavs[-1][-1].value/2])
    # self.bounds = np.asarray([bounds0, bounds1])
    # self.nwav = self.bounds.shape[1]
    # data, grad0, grad1 = np.zeros((3, len(pixs), len(self.locations), *self.shape))
    # rows, cols, datas, grad0s, grad1s = [], [], [], [], []
    # for pix, wav in tqdm(zip(pixs, wavs), total=len(pixs)):
    #     row, col, data, grad0, grad1 = [], [], [], [], []
    #     for location in self.locations * self.psf.scale:
    #         for idx in range(len(pix)):
    #             r, c, ar, g0, g1 = self.psf.prf(
    #                 row=location[0] + pix.value[idx],
    #                 column=location[1],
    #                 wavelength=wav[idx],
    #                 gradients=True,
    #             )

    #         row.append(r[:, None] * np.ones(c.shape[0], int))
    #         col.append(c[None, :] * np.ones(r.shape[0], int)[:, None])
    #         grad0.append(g0 - g0.mean())
    #         grad1.append(g1 - g1.mean())
    #         data.append(ar)
    #     data, row, col, grad0, grad1 = (
    #         np.asarray(data),
    #         np.asarray(row),
    #         np.asarray(col),
    #         np.asarray(grad0),
    #         np.asarray(grad1),
    #     )
    #     rows.append(row)
    #     cols.append(col)
    #     grad0s.append(grad0)
    #     grad1s.append(grad1)
    #     datas.append(data)
    # rows, cols, datas, grad0s, grad1s = (
    #     np.vstack(rows),
    #     np.vstack(cols),
    #     np.vstack(datas),
    #     np.vstack(grad0s),
    #     np.vstack(grad1s),
    # )
    # self.X = SparseWarp3D(
    #     datas.transpose([1, 2, 0]),
    #     rows.transpose([1, 2, 0]) - self.corner[0],
    #     cols.transpose([1, 2, 0]) - self.corner[1],
    #     imshape=self.shape,
    # )
    # self.dX0 = SparseWarp3D(
    #     grad0s.transpose([1, 2, 0]),
    #     rows.transpose([1, 2, 0]) - self.corner[0],
    #     cols.transpose([1, 2, 0]) - self.corner[1],
    #     imshape=self.shape,
    # )
    # self.dX1 = SparseWarp3D(
    #     grad1s.transpose([1, 2, 0]),
    #     rows.transpose([1, 2, 0]) - self.corner[0],
    #     cols.transpose([1, 2, 0]) - self.corner[1],
    #     imshape=self.shape,
    # )
    # return

    # def _get_Xs(self):
    #     rows, cols, datas, grad0s, grad1s = [], [], [], [], []
    #     for pix, wav in tqdm(
    #         zip(self.psf.trace_pixel * self.scale, self.psf.trace_wavelength),
    #         total=len(self.psf.trace_wavelength),
    #     ):
    #         row, col, data, grad0, grad1 = [], [], [], [], []
    #         for location in self.locations * self.scale:
    #             r, c, ar, g0, g1 = self.psf.prf(
    #                 row=location[0] + pix.value,
    #                 column=location[1],
    #                 wavelength=wav,
    #                 gradients=True,
    #             )
    #             row.append(r[:, None] * np.ones(c.shape[0], int))
    #             col.append(c[None, :] * np.ones(r.shape[0], int)[:, None])
    #             grad0.append(g0)
    #             grad1.append(g1)
    #             data.append(ar)
    #         data, row, col, grad0, grad1 = (
    #             np.asarray(data),
    #             np.asarray(row),
    #             np.asarray(col),
    #             np.asarray(grad0),
    #             np.asarray(grad1),
    #         )
    #         rows.append(row)
    #         cols.append(col)
    #         grad0s.append(grad0)
    #         grad1s.append(grad1)
    #         datas.append(data)
    #     rows, cols, datas, grad0s, grad1s = (
    #         np.vstack(rows),
    #         np.vstack(cols),
    #         np.vstack(datas),
    #         np.vstack(grad0s),
    #         np.vstack(grad1s),
    #     )
    #     self.X = SparseWarp3D(
    #         datas.transpose([1, 2, 0]),
    #         rows.transpose([1, 2, 0]) - self.corner[0],
    #         cols.transpose([1, 2, 0]) - self.corner[1],
    #         imshape=self.shape,
    #     )
    #     self.dX0 = SparseWarp3D(
    #         grad0s.transpose([1, 2, 0]),
    #         rows.transpose([1, 2, 0]) - self.corner[0],
    #         cols.transpose([1, 2, 0]) - self.corner[1],
    #         imshape=self.shape,
    #     )
    #     self.dX1 = SparseWarp3D(
    #         grad1s.transpose([1, 2, 0]),
    #         rows.transpose([1, 2, 0]) - self.corner[0],
    #         cols.transpose([1, 2, 0]) - self.corner[1],
    #         imshape=self.shape,
    #     )
    #     return

    def model(
        self,
        spectra: npt.ArrayLike,
        delta_pos: Optional[npt.ArrayLike] = None,
        quiet: bool = False,
        downsample: bool = True,
    ) -> npt.ArrayLike:
        """`spectra` must have shape nwav x ntargets x ntime"""
        if spectra.ndim == 1:
            spectra = spectra[:, None, None]
        elif spectra.ndim == 2:
            spectra = spectra[:, :, None]
        elif spectra.ndim != 3:
            raise ValueError("Pass a 3D array for flux (nwav, ntargets, ntime).")
        if (spectra.shape[0] != self.nwav) | (spectra.shape[1] != len(self)):
            raise ValueError("`spectra` must have shape (nwav, ntargets)")
        if delta_pos is not None:
            delta_pos = deepcopy(delta_pos) * float(self.scale)
            if delta_pos.ndim == 1:
                delta_pos = delta_pos[:, None]
            elif delta_pos.ndim != 2:
                raise ValueError("Pass 2D array for delta_pos (2, ntime).")

        ar = self._get_ar(flux=np.vstack(spectra), delta_pos=delta_pos, quiet=quiet)

        if self.scale == 1:
            return ar
        if downsample:
            return downsample_array(ar, self.scale)
        else:
            return ar



"""Functions to plot up PSFs"""

from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
from IPython.display import HTML
from matplotlib import animation as mplanimation
from matplotlib.animation import FFMpegWriter
from PIL import Image


def _get_v(kwargs, image_type):
    if image_type.lower() == "psf":
        vmin, vmax = kwargs.pop("vmin", 0), kwargs.pop("vmax", 0.001)
    elif image_type.lower() == "prf":
        vmin, vmax = kwargs.pop("vmin", 0), kwargs.pop("vmax", 0.05)
    elif image_type.lower() == "dprf":
        vmin, vmax = kwargs.pop("vmin", -0.05), kwargs.pop("vmax", 0.05)
    elif image_type.lower() == "dpsf":
        vmin, vmax = kwargs.pop("vmin", -0.001), kwargs.pop("vmax", 0.001)
    return vmin, vmax


def plot_spatial(psf, n=3, image_type="PSF", **kwargs):
    """Plot a PSF that has some spatial dimesions"""
    if "row" not in psf.dimension_names:
        raise ValueError("Not a spatial PSF")
    if "column" not in psf.dimension_names:
        raise ValueError("Not a spatial PSF")

    if not (n % 2) == 1:
        n += 1

    locdict = {}
    for dnm in psf.dimension_names:
        if (dnm == "row") | (dnm == "column"):
            continue
        locdict[dnm] = getattr(psf, dnm + "0d")
    vmin, vmax = _get_v(kwargs, image_type)
    cmap = kwargs.pop("cmap", "viridis")
    fig, ax = plt.subplots(n, n, figsize=(n * 2, n * 2))
    for x1, y1 in (
        np.asarray(((np.mgrid[:n, :n] - n // 2) * (600 / (n // 2))))
        .reshape((2, n**2))
        .T
    ):
        jdx = int(x1 // (600 / (n // 2)) + n // 2)
        idx = int(-y1 // (600 / (n // 2)) + n // 2)
        locdict["row"] = x1
        locdict["column"] = y1
        if image_type.lower() == "psf":
            y, x, f = (
                psf.psf_row.value,
                psf.psf_column.value,
                psf.psf(**locdict),
            )
            ax[idx, jdx].set(xticklabels=[], yticklabels=[])
        elif image_type.lower() == "dpsf":
            y, x, df = (
                psf.psf_row.value,
                psf.psf_column.value,
                psf.dpsf(**locdict),
            )
            f = df[0]
            ax[idx, jdx].set(xticklabels=[], yticklabels=[])
        elif image_type.lower() == "prf":
            y, x, f = psf.prf(**locdict)
            ax[idx, jdx].set(xticklabels=[], yticklabels=[])
        elif image_type.lower() == "dprf":
            y, x, df = psf.dprf(**locdict)
            f = df[0]
            ax[idx, jdx].set(xticklabels=[], yticklabels=[])
        else:
            raise ValueError("No such image type. Choose from `'PSF'`.")
        ax[idx, jdx].pcolormesh(
            x,
            y,
            f,
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
        )
        ax[idx, jdx].set(xticks=[], yticks=[])
        if idx == n - 1:
            ax[idx, jdx].set(xlabel=f"{x1.astype(int)} Pix")
        if jdx == 0:
            ax[idx, jdx].set(ylabel=f"{y1.astype(int)} Pix")
    # ax[n // 2, 0].set(ylabel="Y Pixel")
    # ax[n - 1, n // 2].set(xlabel="X Pixel")
    ax[0, n // 2].set(title=image_type.upper())
    plt.subplots_adjust(hspace=0.051, wspace=0.051)
    return fig


def plot_spectral(psf, var="wavelength", n=5, npixels=20, image_type="psf", **kwargs):

    vmin, vmax = _get_v(kwargs, image_type)
    cmap = kwargs.pop("cmap", "viridis")

    wavs = np.linspace(
        getattr(psf, var + "1d").min(), getattr(psf, var + "1d").max(), n
    )
    m = npixels // 2
    fig, ax = plt.subplots(
        1,
        n,
        figsize=(n * 3, 3),
        sharex=True,
        sharey=True,
        facecolor="white",
    )

    kwargs = {}
    for dnm in psf.dimension_names:
        if dnm == var:
            continue
        kwargs[dnm] = getattr(psf, dnm + "0d")

    for ndx in np.arange(n):
        kwargs[var] = wavs[ndx]
        if image_type.lower() == "psf":
            y, x, f = (
                psf.psf_row.value,
                psf.psf_column.value,
                psf.psf(**kwargs),
            )
        elif image_type.lower() == "prf":
            y, x, f = psf.prf(
                row=kwargs.pop("row", 0), column=kwargs.pop("column", 0), **kwargs
            )
        im = ax[ndx].pcolormesh(
            x,
            y,
            f,
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
        )
        ax[ndx].set(
            xlim=(-m, m),
            ylim=(-m, m),
            xticks=np.arange(-(m - 1), m, 2),
            yticks=np.arange(-(m - 1), m, 2),
            title=f"{wavs[ndx]:0.2} $\mu$m",
            xlabel="Pixels",
        )
    #            ax[ndx].grid(True, ls="-", color="white", lw=0.5, alpha=0.5)
    plt.subplots_adjust(wspace=0, hspace=0.2)
    ax[0].set(ylabel="Pixels")
    # for idx in range(n):
    #     ax[idx, 0].set(ylabel="Pixels")
    cbar = plt.colorbar(im, ax=ax)
    cbar.set_label("Normalized Flux Value")
    # fig.suptitle(
    #     f"{image_type.upper()} Across {var.capitalize()}", fontsize=15
    # )
    return fig


def _to_matplotlib_animation(
    data,
    instance_name: str = "Frame",
    step: int = None,
    interval: int = 200,
    position: Optional = None,
    figsize=None,
    **plot_args,
):
    """lifted from Lightkurve"""
    if step is None:
        step = len(data) // 50
        if step < 1:
            step = 1

    if figsize is None:
        x, y = 1.0, 1 * data.shape[1] / data.shape[2]
        if np.max([x, y]) < 10.0:
            c = 10.0 / np.max([x, y])
            x *= c
            y *= c
        x, y = int(x), int(y)
        figsize = (x, y)

    _, ax = plt.subplots(figsize=figsize)
    ax.imshow(data[0], **plot_args)
    ax.set(xticks=[], yticks=[])
    if position is not None:
        plt.gca().set_position(position)

    def init():
        return ax.images

    def animate(i):
        frame = i * step
        ax.images[0].set_data(data[frame])
        ax.set_title(f"{instance_name} {frame}")
        return ax.images

    plt.close(ax.figure)
    frames = len(data) // step
    anim = mplanimation.FuncAnimation(
        ax.figure,
        animate,
        init_func=init,
        frames=frames,
        interval=interval,
        blit=True,
    )
    return anim


def animate(data, step: int = None, interval: int = 200, **plot_args):
    return HTML(
        _to_matplotlib_animation(
            data, step=step, interval=interval, position=[0, 0, 1, 1], **plot_args
        ).to_jshtml()
    )


def save_mp4(
    data,
    outfile="out.mp4",
    step: int = None,
    interval: int = 200,
    instance_name="Frame",
    dpi=100,
    **plot_args,
):
    anim = _to_matplotlib_animation(
        data,
        step=step,
        interval=interval,
        position=[0, 0, 1, 1],
        instance_name=instance_name,
        **plot_args,
    )
    anim.save(outfile, writer=FFMpegWriter(fps=1000 / interval, bitrate=5000), dpi=dpi)


def save_gif(
    data,
    outfile="out.gif",
    step: int = 1,
    duration: int = 50,
    scale: int = 1,
    vmin=-50,
    vmax=50,
):
    imgs = (
        (
            np.min([np.max([data, data**0 + vmin], axis=0), data**0 + vmax], axis=0)
            - vmin
        )
        * 255
        / (vmax - vmin)
    )
    imgs = imgs.astype(np.int32)[::step]
    imgs = np.repeat(np.repeat(imgs, scale, axis=1), scale, axis=2)
    imgs = imgs // 2 * 2
    imgs = [Image.fromarray(img) for img in imgs]
    imgs[0].save(
        outfile, save_all=True, append_images=imgs[1:], duration=duration, loop=0
    )



"""Handy utilities for PSFs"""
from datetime import datetime

import astropy.units as u
import numpy as np
import pandas as pd
from astropy.constants import c, h
from astropy.io import fits
from scipy.io import loadmat
from scipy.ndimage import gaussian_filter

from . import PACKAGEDIR


def make_pixel_files():
    def photon_energy(wavelength):
        return ((h * c) / wavelength) * 1 / u.photon

    def sensitivity(wavelength):
        sed = 1 * u.erg / u.s / u.cm**2 / u.angstrom
        E = photon_energy(wavelength)
        telescope_area = np.pi * (mirror_diameter / 2) ** 2
        photon_flux_density = (telescope_area * sed * throughput(wavelength) / E).to(
            u.photon / u.second / u.angstrom
        ) * qe(wavelength)
        sensitivity = photon_flux_density / sed
        return sensitivity

    def qe(wavelength):
        wavelength = np.atleast_1d(wavelength)
        sw_coeffs = np.array([0.65830, -0.05668, 0.25580, -0.08350])
        sw_exponential = 100.0
        sw_wavecut_red = 1.62  # changed from 2.38 for Pandora
        sw_wavecut_blue = 0.85  # new for Pandora
        with np.errstate(invalid="ignore", over="ignore"):
            sw_qe = (
                sw_coeffs[0]
                + sw_coeffs[1] * wavelength.to(u.micron).value
                + sw_coeffs[2] * wavelength.to(u.micron).value ** 2
                + sw_coeffs[3] * wavelength.to(u.micron).value ** 3
            )

            sw_qe = np.where(
                wavelength.to(u.micron).value > sw_wavecut_red,
                sw_qe
                * np.exp(
                    (sw_wavecut_red - wavelength.to(u.micron).value) * sw_exponential
                ),
                sw_qe,
            )

            sw_qe = np.where(
                wavelength.to(u.micron).value < sw_wavecut_blue,
                sw_qe
                * np.exp(
                    -(sw_wavecut_blue - wavelength.to(u.micron).value)
                    * (sw_exponential / 1.5)
                ),
                sw_qe,
            )
        sw_qe[sw_qe < 1e-5] = 0
        return sw_qe * u.electron / u.photon

    def throughput(wavelength):
        df = pd.read_csv(f"{PACKAGEDIR}/data/dichroic-transmission.csv")
        throughput = np.interp(wavelength.to(u.angstrom).value, *np.asarray(df).T)
        return throughput**2 * 0.71

    mirror_diameter = 0.43 * u.m

    df = pd.read_csv(f"{PACKAGEDIR}/data/pixel_vs_wavelength.csv")
    pixel = np.round(np.arange(-400, 80, 0.25), 5) * u.pixel
    wav = np.polyval(np.polyfit(df.Pixel, df.Wavelength, 3), pixel.value) * u.micron

    sens = sensitivity(wav)
    mask = sens / sens.max() > 0.0001
    corr = np.trapz(sens, wav)
    hdu = fits.TableHDU.from_columns(
        [
            fits.Column(
                name="pixel",
                format="D",
                array=pixel.value[mask],
                unit=pixel.unit.to_string(),
            ),
            fits.Column(
                name="wavelength",
                format="D",
                array=wav.value[mask],
                unit=wav.unit.to_string(),
            ),
            fits.Column(
                name="sensitivity",
                format="D",
                array=(sens[mask] / corr),
                unit=(sens / corr).unit.to_string(),
            ),
        ]
    )
    hdu.header.append(
        fits.Card("SENSCORR", corr.value, "correction to apply to sensitivity")
    )
    hdu.header.append(
        fits.Card("CORRUNIT", corr.unit.to_string(), "units of correction")
    )
    hdu.writeto(f"{PACKAGEDIR}/data/nirda-wav-solution.fits", overwrite=True)

    def qe(wavelength):  # noqa: F811
        df = pd.read_csv(f"{PACKAGEDIR}/data/pandora_visible_qe.csv")
        wav, transmission = np.asarray(df.Wavelength) * u.angstrom, np.asarray(
            df.Transmission
        )
        return (
            np.interp(wavelength, wav, transmission, left=0, right=0)
            * u.electron
            / u.photon
        )

    def throughput(wavelength):  # noqa: F811
        df = pd.read_csv(f"{PACKAGEDIR}/data/dichroic-transmission.csv")
        throughput = 1 - np.interp(wavelength.to(u.angstrom).value, *np.asarray(df).T)
        return throughput * 0.752

    wav = np.arange(0.25, 1.3, 0.01) * u.micron
    pixel = np.zeros(len(wav)) * u.pixel
    sens = sensitivity(wav)
    corr = np.trapz(sens, wav)

    hdu = fits.TableHDU.from_columns(
        [
            fits.Column(
                name="pixel", format="D", array=pixel.value, unit=pixel.unit.to_string()
            ),
            fits.Column(
                name="wavelength",
                format="D",
                array=wav.value,
                unit=wav.unit.to_string(),
            ),
            fits.Column(
                name="sensitivity",
                format="D",
                array=(sens / corr),
                unit=(sens / corr).unit.to_string(),
            ),
        ]
    )
    hdu.header.append(
        fits.Card("SENSCORR", corr.value, "correction to apply to sensitivity")
    )
    hdu.header.append(
        fits.Card("CORRUNIT", corr.unit.to_string(), "units of correction")
    )
    hdu.writeto(f"{PACKAGEDIR}/data/visda-wav-solution.fits", overwrite=True)


def _open_LLNL_file(filename, pixel_size, downsample=None, trim=None):
    """This opens the newest file format from LLNL. They might change it, so watch out."""
    d = loadmat(filename)
    subpixel_size = d["dx"][0][0] * u.micron / u.pix

    wvl = np.unique(d["wvl"]) * u.micron
    nwav = wvl.shape[0]
    #    npoints = int(d["x"].shape[1] / nwav)

    TC, TR = np.meshgrid(
        np.unique(d["Tx"][0, ::nwav]), np.unique(d["Ty"][0, ::nwav]), indexing="xy"
    )
    sC, sR = np.meshgrid(
        np.arange(d["PSF"].shape[0]), np.arange(d["PSF"].shape[1]), indexing="xy"
    )
    mask = np.any(
        [
            (TR == row) & (TC == col)
            for col, row in zip(d["Tx"][0, ::nwav], d["Ty"][0, ::nwav])
        ],
        axis=0,
    )
    C, R = np.zeros((2, nwav, *TR.shape))

    for idx in range(nwav):
        C[idx, mask] = d["x"][0][idx::nwav]
        R[idx, mask] = d["y"][0][idx::nwav]
    row1d, column1d = R[:, C.shape[2] // 2].mean(axis=0), C[:, :, C.shape[2] // 2].mean(
        axis=0
    )

    PSF = np.zeros((*d["PSF"].shape[:2], *TR.shape, wvl.shape[0]))

    for wdx in np.arange(nwav):
        PSF[:, :, mask, wdx] = d["PSF"][:, :, wdx::nwav]
    # We need to sort of extrapolate the edges because we don't have those values.
    for count in range(2):
        extrap = np.zeros(PSF.shape)
        extrap[:, :, 1:, :, :] += PSF[:, :, :-1, :, :]
        extrap[:, :, :-1, :, :] += PSF[:, :, 1:, :, :]
        extrap[:, :, :, 1:, :] += PSF[:, :, :, :-1, :]
        extrap[:, :, :, :-1, :] += PSF[:, :, :, 1:, :]
        extrap /= 2
        PSF[:, :, ~mask, :] += extrap[:, :, ~mask, :]

    PSF /= PSF.sum(axis=(0, 1))[None, None]

    # Gaussian blur with std dev of 0.5 detector pixels
    blur_width = ((pixel_size * 0.5 * u.pix) / subpixel_size).value

    PSF = np.asarray(
        [
            [
                [
                    gaussian_filter(PSF[:, :, idx, jdx, kdx], blur_width)
                    for idx in range(PSF.shape[2])
                ]
                for jdx in range(PSF.shape[3])
            ]
            for kdx in range(PSF.shape[4])
        ]
    )

    PSF = PSF.transpose([4, 3, 2, 1, 0])
    PSF = PSF.transpose([1, 0, 3, 2, 4])
    row = row1d[:, None, None] * np.ones((*TR.shape, nwav)) * u.mm
    column = column1d[None, :, None] * np.ones((*TR.shape, nwav)) * u.mm
    wavelength = wvl[None, None, :] * np.ones((*TR.shape, nwav))
    subpixel_row = ((sR[:, 0] - sR.shape[0] / 2) * u.pixel * subpixel_size)[
        :, None
    ] * np.ones(d["PSF"].shape[:2])
    subpixel_column = ((sC[0] - sC.shape[0] / 2) * u.pixel * subpixel_size)[
        None
    ] * np.ones(d["PSF"].shape[:2])

    if downsample is not None:
        n = downsample
        PSF = np.sum(
            [PSF[idx::n, jdx::n] for idx in range(n) for jdx in range(n)], axis=0
        )
        subpixel_column = (
            np.mean(
                [
                    subpixel_column.value[idx::n, jdx::n]
                    for idx in range(n)
                    for jdx in range(n)
                ],
                axis=0,
            )
            * subpixel_column.unit
        )
        subpixel_row = (
            np.mean(
                [
                    subpixel_row.value[idx::n, jdx::n]
                    for idx in range(n)
                    for jdx in range(n)
                ],
                axis=0,
            )
            * subpixel_row.unit
        )
        subpixel_size *= downsample

    if trim is not None:
        PSF = PSF[trim:-trim, trim:-trim]
        subpixel_column = subpixel_column[trim:-trim, trim:-trim]
        subpixel_row = subpixel_row[trim:-trim, trim:-trim]

    return (
        subpixel_size,
        subpixel_row,
        subpixel_column,
        row,
        column,
        wavelength,
        PSF,
        str(d["__header__"]).split("Created on: ")[-1][:-1],
    )


def make_PSF_fits_files(filename, suffix, pixel_size, downsample=None, trim=None):
    (
        subpixel_size,
        subpixel_row,
        subpixel_column,
        row,
        column,
        wavelength,
        PSF,
        created_on,
    ) = _open_LLNL_file(filename, pixel_size, downsample=downsample, trim=trim)
    pixel_size = u.Quantity(pixel_size, u.micron / u.pixel)
    subpixel_row, subpixel_column, row, column = (
        (subpixel_row / pixel_size).to(u.pixel),
        (subpixel_column / pixel_size).to(u.pixel),
        (row / pixel_size).to(u.pixel),
        (column / pixel_size).to(u.pixel),
    )

    hdr = fits.Header(
        [
            ("AUTHOR1", "LLNL"),
            ("AUTHOR2", "Pandora DPC"),
            ("ORIGIN", filename.split("/")[-1]),
            ("SUFFIX", suffix),
            ("CREATED", created_on),
            ("DATE", datetime.now().isoformat()),
            (
                "PIXSIZE",
                pixel_size.value,
                f"PSF pixel size in {pixel_size.unit.to_string()}",
            ),
            (
                "SUBPIXSZ",
                subpixel_size.value,
                f"PSF sub pixel size in {subpixel_size.unit.to_string()}",
            ),
        ]
    )
    primaryhdu = fits.PrimaryHDU(header=hdr)
    hdu = fits.HDUList(
        [
            primaryhdu,
            fits.ImageHDU(PSF.astype(np.float32), name="PSF"),
            fits.ImageHDU(row.value.astype(np.float32), name="ROW"),
            fits.ImageHDU(column.value.astype(np.float32), name="COLUMN"),
            fits.ImageHDU(wavelength.value.astype(np.float32), name="WAVELENGTH"),
        ]
    )
    hdu[2].header["UNIT"] = row.unit.to_string()
    hdu[3].header["UNIT"] = column.unit.to_string()
    hdu[4].header["UNIT"] = wavelength.unit.to_string()
    hdu.writeto(
        PACKAGEDIR
        + f"/data/pandora_{suffix}_{datetime.strftime(datetime.now(), '%Y-%m')}.fits",
        overwrite=True,
    )


# def make_vis_PSF_fits_files(dir, nbin=1, suffix="_hr"):
#     d = loadmat(dir + "pandora_vis_20220506_hot_PSF_512.mat")

#     PSF_hot = d["PSF"][:-1, :-1, :]  # [128:384, 128:384]
#     PSF_hot = np.asarray(
#         [[PSF_hot[idx::nbin, jdx::nbin] for idx in range(nbin)] for jdx in range(nbin)]
#     ).mean(axis=(0, 1))
#     PSF_hot = PSF_hot.reshape((*PSF_hot.shape[:2], 9, 9, 5))
#     d = loadmat(dir + "pandora_vis_20220506_cold_PSF_512.mat")

#     PSF_cold = d["PSF"][:-1, :-1, :]  # [128:384, 128:384]
#     PSF_cold = np.asarray(
#         [[PSF_cold[idx::nbin, jdx::nbin] for idx in range(nbin)] for jdx in range(nbin)]
#     ).mean(axis=(0, 1))
#     PSF_cold = PSF_cold.reshape((*PSF_cold.shape[:2], 9, 9, 5))

#     sub_pixel_size = nbin * d["dx"][0][0] * u.micron / u.pix
#     pixel_size = 6.5 * u.micron / u.pix
#     x = d["x"][0].reshape((9, 9, 5)) * u.mm
#     y = d["y"][0].reshape((9, 9, 5)) * u.mm
#     x = (x / pixel_size).to(u.pixel)
#     y = (y / pixel_size).to(u.pixel)
#     wvl = d["wvl"][0].reshape((9, 9, 5)) * u.micron
#     temp = np.asarray([-10.0, 30.0]) * u.deg_C
#     PSF = np.asarray([PSF_cold, PSF_hot]).transpose([1, 2, 3, 4, 5, 0])
#     PSF /= PSF.sum(axis=(0, 1))[None, None]

#     PSF = np.asarray(
#         [
#             [
#                 [
#                     [
#                         gaussian_filter(PSF[:, :, idx, jdx, kdx, ldx], 6)
#                         for idx in range(PSF.shape[2])
#                     ]
#                     for jdx in range(PSF.shape[3])
#                 ]
#                 for kdx in range(PSF.shape[4])
#             ]
#             for ldx in range(PSF.shape[5])
#         ]
#     )
#     PSF = PSF.transpose([4, 5, 3, 2, 1, 0])
#     x = x[:, :, :, None] * np.ones(len(temp))
#     y = y[:, :, :, None] * np.ones(len(temp))
#     wvl = wvl[:, :, :, None] * np.ones(len(temp))
#     temp = temp[None, None, None, :] * np.ones(wvl.shape)

#     hdr = fits.Header(
#         [
#             ("AUTHOR1", "LLNL"),
#             ("AUTHOR2", "Christina Hedges"),
#             ("ORIGIN1", "pandora_vis_20220506_cold_PSF_512.mat"),
#             ("ORIGIN2", "pandora_vis_20220506_hot_PSF_512.mat"),
#             ("CREATED", str(d["__header__"]).split("Created on: ")[-1][:-1]),
#             ("DATE", datetime.now().isoformat()),
#             (
#                 "PIXSIZE",
#                 pixel_size.value,
#                 f"PSF pixel size in {pixel_size.unit.to_string()}",
#             ),
#             (
#                 "SUBPIXSZ",
#                 sub_pixel_size.value,
#                 f"PSF sub pixel size in {sub_pixel_size.unit.to_string()}",
#             ),
#         ]
#     )
#     primaryhdu = fits.PrimaryHDU(header=hdr)
#     hdu = fits.HDUList(
#         [
#             primaryhdu,
#             fits.ImageHDU(PSF, name="PSF"),
#             fits.ImageHDU(x.value, name="X"),
#             fits.ImageHDU(y.value, name="Y"),
#             fits.ImageHDU(wvl.value, name="WAVELENGTH"),
#             fits.ImageHDU(temp.value, name="TEMPERATURE"),
#         ]
#     )
#     hdu[2].header["UNIT"] = y.unit.to_string()
#     hdu[3].header["UNIT"] = x.unit.to_string()
#     hdu[4].header["UNIT"] = wvl.unit.to_string()
#     hdu[5].header["UNIT"] = temp.unit.to_string()
#     hdu.writeto(
#         PACKAGEDIR + f"/data/pandora_vis{suffix}_20220506.fits",
#         overwrite=True,
#     )


# def make_nir_PSF_fits_files(dir, nbin=1, suffix="_hr"):
#     # NIR PSF with thermal info
#     # -----------------------------------#

#     d = loadmat(dir + "pandora_nir_20220506_thin_prism_hot_PSF_512.mat")
#     PSF_hot = d["PSF"][:-1, :-1, :-1][128:384, 128:384]
#     PSF_hot = np.asarray(
#         [[PSF_hot[idx::nbin, jdx::nbin] for idx in range(nbin)] for jdx in range(nbin)]
#     ).mean(axis=(0, 1))

#     d = loadmat(dir + "pandora_nir_20220506_thin_prism_cold_PSF_512.mat")

#     PSF_cold = d["PSF"][:-1, :-1, :-1][128:384, 128:384]
#     PSF_cold = np.asarray(
#         [[PSF_cold[idx::nbin, jdx::nbin] for idx in range(nbin)] for jdx in range(nbin)]
#     ).mean(axis=(0, 1))

#     sub_pixel_size = nbin * d["dx"][0][0] * u.micron / u.pix
#     pixel_size = 18 * u.micron / u.pix

#     # Bin down wavelength too, also too big.
#     wbin = 10
#     PSF_cold = np.asarray([PSF_cold[:, :, idx::wbin] for idx in range(wbin)]).mean(
#         axis=(0)
#     )
#     PSF_hot = np.asarray([PSF_hot[:, :, idx::wbin] for idx in range(wbin)]).mean(
#         axis=(0)
#     )
#     wvl = (
#         np.asarray([d["wvl"][:900][idx::wbin] for idx in range(wbin)]).mean(axis=(0))[
#             :, 0
#         ]
#         * u.micron
#     )
#     temp = np.asarray([-10.0, 30.0]) * u.deg_C
#     wvl = wvl[:, None] * np.ones(len(temp))
#     temp = temp[None, :] * np.ones(wvl.shape)

#     PSF = np.asarray([PSF_cold, PSF_hot]).transpose([1, 2, 3, 0])
#     PSF /= PSF.sum(axis=(0, 1))

#     PSF = np.asarray(
#         [
#             [gaussian_filter(PSF[:, :, idx, jdx], 3) for idx in range(PSF.shape[2])]
#             for jdx in range(PSF.shape[3])
#         ]
#     )
#     PSF = PSF.transpose([2, 3, 1, 0])

#     hdr = fits.Header(
#         [
#             ("AUTHOR1", "LLNL"),
#             ("AUTHOR2", "Christina Hedges"),
#             ("ORIGIN", "pandora_nir_20220506.mat"),
#             ("CREATED", str(d["__header__"]).split("Created on: ")[-1][:-1]),
#             ("DATE", datetime.now().isoformat()),
#             (
#                 "PIXSIZE",
#                 pixel_size.value,
#                 f"PSF pixel size in {pixel_size.unit.to_string()}",
#             ),
#             (
#                 "SUBPIXSZ",
#                 sub_pixel_size.value,
#                 f"PSF sub pixel size in {sub_pixel_size.unit.to_string()}",
#             ),
#         ]
#     )
#     primaryhdu = fits.PrimaryHDU(header=hdr)
#     hdu = fits.HDUList(
#         [
#             primaryhdu,
#             fits.ImageHDU(PSF, name="PSF"),
#             fits.ImageHDU(wvl.value, name="WAVELENGTH"),
#             fits.ImageHDU(temp.value, name="TEMPERATURE"),
#         ]
#     )
#     hdu[2].header["UNIT"] = wvl.unit.to_string()
#     hdu[3].header["UNIT"] = temp.unit.to_string()
#     hdu.writeto(
#         PACKAGEDIR + f"/data/pandora_nir{suffix}_20220506.fits",
#         overwrite=True,
#     )


# def make_gaussian_psf():
#     hdu1 = fits.open(f"{PACKAGEDIR}/data/pandora_vis_hr_20220506.fits")
#     pixel_size = hdu1[0].header["PIXSIZE"] * u.micron / u.pix
#     sub_pixel_size = hdu1[0].header["SUBPIXSZ"] * u.micron / u.pix
#     shape = hdu1[1].data.shape[:2]
#     psf_column = (
#         (np.arange(shape[1]) - shape[1] // 2) * u.pixel * sub_pixel_size
#     ) / pixel_size
#     psf_row = (
#         (np.arange(shape[0]) - shape[0] // 2) * u.pixel * sub_pixel_size
#     ) / pixel_size

#     def gauss(row, col, sigma):
#         return (1 / (sigma * np.sqrt(2 * np.pi))) * np.exp(
#             -(
#                 (col[None, :]) ** 2 / (2 * sigma**2)
#                 + (row[:, None]) ** 2 / (2 * sigma**2)
#             )
#         )

#     R = np.abs(hdu1[2].data[:, 0, 0, 0]) / 600 + 1
#     C = np.abs(hdu1[3].data[0, :, 0, 0]) / 600 + 1

#     ar = np.zeros((9, 9, 256, 256))
#     for idx, r1 in enumerate(R):
#         for jdx, c1 in enumerate(C):
#             ar[idx, jdx] = gauss(psf_row.value, psf_column.value, np.hypot(r1, c1))
#     ar = ar.transpose([2, 3, 0, 1])
#     integral = np.trapz(
#         np.trapz(ar, dx=np.median(np.diff(psf_row).value), axis=0),
#         dx=np.median(np.diff(psf_column).value),
#         axis=0,
#     )
#     ar /= integral[None, None, :, :]
#     hdr = fits.Header(
#         [
#             ("AUTHOR1", "Christina Hedges"),
#             ("ORIGIN1", "Gaussian"),
#             ("CREATED", datetime.now().isoformat()),
#             ("DATE", datetime.now().isoformat()),
#             (
#                 "PIXSIZE",
#                 pixel_size.value,
#                 f"PSF pixel size in {pixel_size.unit.to_string()}",
#             ),
#             (
#                 "SUBPIXSZ",
#                 sub_pixel_size.value,
#                 f"PSF sub pixel size in {sub_pixel_size.unit.to_string()}",
#             ),
#         ]
#     )
#     primaryhdu = fits.PrimaryHDU(header=hdr)
#     hdu = fits.HDUList(
#         [
#             primaryhdu,
#             fits.ImageHDU(ar, name="PSF"),
#             fits.ImageHDU(hdu1[2].data[:, :, 0, 0], name="X"),
#             fits.ImageHDU(hdu1[3].data[:, :, 0, 0], name="Y"),
#         ]
#     )
#     hdu[2].header["UNIT"] = u.pixel.to_string()
#     hdu[3].header["UNIT"] = u.pixel.to_string()
#     hdu.writeto(
#         f"{PACKAGEDIR}/data/pandora_gaussian_psf.fits",
#         overwrite=True,
#     )


def prep_for_add(row, column, prf, shape=(100, 100), corner=(-50, -50)):
    Y, X = np.asarray(
        np.meshgrid(
            row - corner[0],
            column - corner[1],
            indexing="ij",
        )
    ).astype(int)
    k = (X >= 0) & (X < shape[1]) & (Y >= 0) & (Y < shape[0])
    if prf.ndim == 2:
        return Y[k], X[k], prf[k]
    elif prf.ndim == 3:
        return Y[k], X[k], prf[:, k]
    else:
        raise ValueError("can not parse prf for adding")


# def bin_prf(psf0, psf_column, psf_row, location=(0, 0), normalize=True):
#     mod = (psf_column + location[1]) % 1
#     cyc = ((psf_column + location[1]) - mod).astype(int)
#     colbin = np.unique(cyc)
#     psf1 = np.asarray(
#         [psf0[:, cyc == c].sum(axis=1) / (cyc == c).sum() for c in colbin]
#     ).T

#     mod = (psf_row + location[0]) % 1
#     cyc = ((psf_row + location[0]) - mod).astype(int)
#     rowbin = np.unique(cyc)
#     psf2 = np.asarray([psf1[cyc == c].sum(axis=0) / (cyc == c).sum() for c in rowbin])
#     if normalize:
#         psf2 /= psf2.sum()
#     return rowbin.astype(int), colbin.astype(int), psf2


def bin_prf(psf0, psf_column, psf_row, location=(0, 0), normalize=True):
    df = pd.DataFrame(
        psf0, index=psf_row + location[0], columns=psf_column + location[1]
    )
    int_location = (
        int(location[0] - (location[0] % 1)),
        int(location[1] - (location[1] % 1)),
    )
    irange, crange = (
        np.floor(psf_row.min()).astype(int) - 1 + int_location[0],
        np.ceil(psf_row.max()).astype(int) + 2 + int_location[0],
    ), (
        np.floor(psf_column.min()).astype(int) - 1 + int_location[1],
        np.ceil(psf_column.max()).astype(int) + 2 + int_location[1],
    )
    df2 = (
        df.groupby(by=df.index - (df.index % 1))
        .sum()
        .T.groupby(by=df.columns - (df.columns % 1))
        .sum()
        .reindex(range(*crange), fill_value=0)
        .T.reindex(range(*irange), fill_value=0)
    )
    if normalize:
        df2 /= df2.sum().sum()
    #    import pdb;pdb.set_trace()
    return np.asarray(list(df2.index)), np.asarray(list(df2.columns)), df2.values


def downsample(ar, n=2):
    if not ar.ndim == 3:
        raise ValueError("Pass a 3D array")
    if not isinstance(n, int):
        raise ValueError("Can only downsample by an integer")

    # Calculate the dimensions for the result array
    new_shape = (ar.shape[0], *tuple(np.asarray(ar.shape[1:]) // n))

    # Reshape the input array into smaller blocks
    try:
        reshaped_array = ar.reshape(
            ar.shape[0],
            new_shape[1],
            ar.shape[1] // new_shape[1],
            new_shape[2],
            ar.shape[2] // new_shape[2],
        )
    except ValueError:
        raise ValueError(
            f"Can not reshape array shape {ar.shape} into array shape {new_shape}."
        )
    # Take the mean over the new dimensions
    return reshaped_array.sum(axis=(2, 4))



"""Magic classes for working with sparse data."""
from copy import deepcopy
from typing import Tuple

import numpy as np
from scipy import sparse

__all__ = ["SparseWarp3D", "ROISparseWarp3D"]


class SparseWarp3D(sparse.coo_matrix):
    """Special class for working with stacks of sparse 3D images"""

    def __init__(self, data, row, col, imshape):
        if not np.all([row.ndim == 3, col.ndim == 3, data.ndim == 3]):
            raise ValueError("Pass a 3D array (nrow, ncol, nvecs)")
        self.nvecs = data.shape[-1]
        if not np.all(
            [
                row.shape[-1] == self.nvecs,
                col.shape[-1] == self.nvecs,
            ]
        ):
            raise ValueError("Must have the same 3rd dimension (nvecs).")
        self.subrow = row.astype(int)
        self.subcol = col.astype(int)
        self.subdepth = (
            np.arange(row.shape[-1], dtype=int)[None, None, :]
            * np.ones(row.shape, dtype=int)[:, :, None]
        )
        self.subdata = data
        self._kz = self.subdata != 0

        self.imshape = imshape
        self.subshape = row.shape
        self.cooshape = (np.prod([*self.imshape[:2]]), self.nvecs)
        self.coord = (0, 0)
        super().__init__(self.cooshape)
        index0 = (np.vstack(self.subrow)) * self.imshape[1] + (np.vstack(self.subcol))
        index1 = np.vstack(self.subdepth).ravel()
        self._index_no_offset = np.vstack([index0.ravel(), index1.ravel()])
        self._submask_no_offset = np.vstack(self._get_submask(offset=(0, 0))).ravel()
        self._subrow_v = deepcopy(np.vstack(self.subrow).ravel())
        self._subcol_v = deepcopy(np.vstack(self.subcol).ravel())
        self._subdata_v = deepcopy(np.vstack(deepcopy(self.subdata)).ravel())
        self._index1 = np.vstack(self.subdepth).ravel()

        self._set_data()

    def __add__(self, other):
        if isinstance(other, SparseWarp3D):
            data = deepcopy(self.subdata + other.subdata)
            if (
                (self.subcol != other.subcol)
                | (self.subrow != other.subrow)
                | (self.imshape != other.imshape)
                | (self.subshape != other.subshape)
            ):
                raise ValueError("Must have same base indicies.")
            return SparseWarp3D(
                data=data, row=self.subrow, col=self.subcol, imshape=self.imshape
            )
        else:
            return super(sparse.coo_matrix, self).__add__(other)

    def tocoo(self):
        return sparse.coo_matrix((self.data, (self.row, self.col)), shape=self.cooshape)

    def index(self, offset=(0, 0)):
        """Get the 2D positions of the data"""
        if offset == (0, 0):
            return self._index_no_offset
        index0 = (self._subrow_v + offset[0]) * self.imshape[1] + (
            self._subcol_v + offset[1]
        )
        #        index1 = np.vstack(self.subdepth).ravel()
        #        return np.vstack([index0.ravel(), index1.ravel()])
        # index0 = (self._subrow_v + offset[0]) * self.imshape[1] + (
        #     self._subcol_v * offset[1]
        # )
        return index0, self._index1
        # index0 = (self._subrow_v + offset[0]) * self.imshape[1] + (
        #     self._subcol_v * offset[1]
        # )
        # return index0, self._index1

    def _get_submask(self, offset=(0, 0)):
        # find where the data is within the array bounds
        kr = ((self.subrow + offset[0]) < self.imshape[0]) & (
            (self.subrow + offset[0]) >= 0
        )
        kc = ((self.subcol + offset[1]) < self.imshape[1]) & (
            (self.subcol + offset[1]) >= 0
        )
        return kr & kc & self._kz

    def _set_data(self, offset=(0, 0)):
        if offset == (0, 0):
            index0, index1 = self.index((0, 0))
            self.row, self.col = (
                index0[self._submask_no_offset],
                index1[self._submask_no_offset],
            )
            self.data = self._subdata_v[self._submask_no_offset]
        else:
            # find where the data is within the array bounds
            k = self._get_submask(offset=offset)
            k = np.vstack(k).ravel()
            new_row, new_col = self.index(offset=offset)
            self.row, self.col = new_row[k], new_col[k]
            self.data = self._subdata_v[k]
        self.coord = offset

    def __repr__(self):
        return (
            f"<{(*self.imshape, self.nvecs)} SparseWarp3D array of type {self.dtype}>"
        )

    def dot(self, other):
        if not isinstance(other, np.ndarray):
            raise NotImplementedError(
                f"dot products with type {type(other)} are not implemented."
            )
        if other.ndim == 1:
            other = other[:, None]
        nt = other.shape[1]
        return super().dot(other).reshape((*self.imshape, nt)).transpose([2, 0, 1])

    def reset(self):
        """Reset any translation back to the original data"""
        self._set_data(offset=(0, 0))
        self.coord = (0, 0)
        return

    def clear(self):
        """Clear data in the array"""
        self.data = np.asarray([])
        self.row = np.asarray([])
        self.col = np.asarray([])
        self.coord = (0, 0)
        return

    def translate(self, position: Tuple):
        """Translate the data in the array by `position` in (row, column)"""
        self.reset()
        # If translating to (0, 0), do nothing
        if position == (0, 0):
            return
        self.clear()
        self._set_data(position)
        return

    def copy(self):
        return deepcopy(self)


class ROISparseWarp3D(SparseWarp3D):
    """Special version of a SparseWarp3D matrix which only populates and works with data within Regions of Interest."""

    def __init__(self, data, row, col, imshape, nROIs, ROI_size, ROI_corners):
        self.nROIs = nROIs
        self.ROI_size = ROI_size
        self.ROI_corners = ROI_corners
        self.get_ROI_mask = self.parse_ROIS(nROIs, ROI_size, ROI_corners)
        super().__init__(data=data, row=row, col=col, imshape=imshape)

    def parse_ROIS(self, nROIs: int, ROI_size: tuple, ROI_corners: list):
        if not len(ROI_corners) == nROIs:
            raise ValueError("Must pass corners for all ROIs.")
        if not np.all([isinstance(corner, tuple) for corner in ROI_corners]):
            raise ValueError("Pass corners as tuples.")

        def get_ROI_masks(row, column):
            mask = []
            for roi in range(nROIs):
                rmin, cmin = ROI_corners[roi]
                rmax, cmax = rmin + ROI_size[0], cmin + ROI_size[1]
                mask.append(
                    (row >= rmin) & (row < rmax) & (column >= cmin) & (column < cmax)
                )
            return np.asarray(mask)

        return get_ROI_masks

    def __repr__(self):
        return f"<{(*self.imshape, self.nvecs)} ROISparseWarp3D array of type {self.dtype}, {self.nROIs} Regions of Interest>"

    def _get_submask(self, offset=(0, 0)):
        # find where the data is within the array bounds
        kr = ((self.subrow + offset[0]) < self.imshape[0]) & (
            (self.subrow + offset[0]) >= 0
        )
        kc = ((self.subcol + offset[1]) < self.imshape[1]) & (
            (self.subcol + offset[1]) >= 0
        )
        # kroi = self.get_ROI_mask(self.subrow + offset[0], self.subcol + offset[0]).any(
        #     axis=0
        # )
        return kr & kc & self._kz  # & kroi

    def dot(self, other):
        if isinstance(other, np.ndarray):
            other = sparse.csr_matrix(other).T
        if not sparse.issparse(other):
            raise ValueError("Must pass a `sparse` array to dot.")
        if not other.shape[0] == self.nvecs:
            if other.shape[1] == self.nvecs:
                other = other.T
            else:
                raise ValueError(f"Must pass {(self.nvecs, 1)} shape object.")
        sparse_array = super().tocsr().dot(other)

        R, C = np.meshgrid(
            np.arange(0, self.ROI_size[0]),
            np.arange(0, self.ROI_size[1]),
            indexing="ij",
        )
        array = np.zeros((self.nROIs, other.shape[1], *self.ROI_size))
        for rdx, c in enumerate(self.ROI_corners):
            idx = (R.ravel() + c[0]) * self.imshape[1] + (C.ravel() + c[1])
            k = (idx >= 0) & (idx < self.shape[0])
            array[rdx, :, k.reshape(self.ROI_size)] = sparse_array[
                idx[k]
            ].toarray()  # ).reshape(self.ROI_size))
        return array



__version__ = "0.2.9"
# Standard library
import os  # noqa

PACKAGEDIR = os.path.abspath(os.path.dirname(__file__))
TESTDIR = "/".join(PACKAGEDIR.split("/")[:-2]) + "/tests/"
PANDORASTYLE = "{}/data/pandora.mplstyle".format(PACKAGEDIR)

# Standard library
import logging  # noqa: E402
import shutil  # noqa: E402

# Third-party
from rich.logging import RichHandler  # noqa: E402

logging.basicConfig()
log = logging.getLogger("pandora-psf")
log.addHandler(RichHandler(markup=True))
log.setLevel("INFO")

if not os.path.isfile(f"{PACKAGEDIR}/data/pandora_vis_2024-05.fits"):
    from astropy.utils.data import download_file  # noqa: E402

    # Download vis PSF
    log.warning("No PSF file found. Downloading 2GB VIS PSF file.")
    p = download_file(
        "https://zenodo.org/records/11153153/files/pandora_vis_2024-05.fits?download=1",
        pkgname="pandora-psf",
    )
    shutil.move(p, f"{PACKAGEDIR}/data/pandora_vis_2024-05.fits")
    log.warning(f"VIS PSF downloaded to {PACKAGEDIR}/data/pandora_vis_2024-05.fits.")

if not os.path.isfile(f"{PACKAGEDIR}/data/pandora_nir_2024-05.fits"):
    from astropy.utils.data import download_file  # noqa: E402

    # Download nir PSF
    log.warning("No PSF file found. Downloading 2GB NIR PSF")
    p = download_file(
        "https://zenodo.org/records/11153153/files/pandora_nir_2024-05.fits?download=1",
        pkgname="pandora-psf",
    )
    shutil.move(p, f"{PACKAGEDIR}/data/pandora_nir_2024-05.fits")
    log.warning(f"NIR PSF downloaded to {PACKAGEDIR}/data/pandora_nir_2024-05.fits.")

from .psf import PSF  # noqa: F401, E402
from .scene import ROIScene, Scene, TraceScene  # noqa: F401, E402
from .sparsewarp import ROISparseWarp3D, SparseWarp3D  # noqa: F401, E402



"""Defines the PSF class"""

# Standard library
from typing import Dict, List, Union

# Third-party
import astropy.units as u
import numpy as np
import numpy.typing as npt
import pandorasat as ps
from astropy.io import fits
from pandorasat.utils import get_phoenix_model

from . import PACKAGEDIR
from .utils import bin_prf

__all__ = ["PSF"]


class PSF(object):
    """Class to use PSFs"""

    def __init__(
        self,
        name: str,
        X: npt.ArrayLike,
        psf_flux: npt.ArrayLike,
        dimension_names: List,
        dimension_units: List,
        pixel_size: u.Quantity,
        sub_pixel_size: u.Quantity,
        transpose: bool = False,
        freeze_dictionary: Dict = {},
        #        blur_value: Tuple = (0 * u.pixel, 0 * u.pixel),
        check_bounds: bool = True,
        extrapolate: bool = False,
        scale: int = 1,
        bin: int = 1,
    ):
        """
        PSF class for making PSFs, PRFs, and traces.

        Parameters:
        -----------
        X : np.NDArray
            Array of 1D vectors defining the value of each dimension for every element of `psf_flux`.
            Should have as many entries as `psf_flux` has dimensions.
        psf_flux: np.NDArray
            ND array of flux values
        dimension_names: list
            List of names for each of the N dimensions in `psf_flux`
        dimension_units: list
            List of `astropy.unit.Quantity`'s describing units of each dimension
        pixel_size: u.Quantity
            True detector pixel size in dimensions of length/pixel
        sub_pixel_size: u.Quantity
            PSF file pixel size in dimensions of length/pixel
        transpose: bool
            Whether to transpose the data in the column/row axis
        freeze_dictionary: dict
            Dictionary of dimensions to freeze
        blur_value: tuple
            Tuple of astropy quantities in pixels describing the amount of blur in each axis.
        extrapolate : bool
            Whether to allow the PSF to be evaluated outside of the bounds (i.e. will extrapolate)
        """
        if bin == 1:
            self.psf_flux = psf_flux
        elif bin >= 1:
            self.psf_flux = np.asarray(
                [
                    [psf_flux[idx::bin, jdx::bin] for idx in range(bin)]
                    for jdx in range(bin)
                ]
            ).mean(axis=(0, 1))
        else:
            raise ValueError("Can not parse `bin`")
        self.name = name
        self.bin = bin
        self.transpose = transpose
        if self.transpose:
            self.psf_flux = self.psf_flux.transpose(
                [1, 0, *np.arange(2, self.psf_flux.ndim)]
            )
        self.dimension_names = dimension_names
        self.dimension_units = dimension_units
        self.pixel_size = pixel_size
        self.sub_pixel_size = sub_pixel_size * bin
        self.freeze_dictionary = freeze_dictionary
        #        self.blur_value = blur_value
        self.check_bounds = check_bounds
        self.extrapolate = extrapolate
        self.scale = scale

        for name, x in zip(dimension_names, X):
            setattr(self, name, x)

        self._validate()
        self.psf_flux_grad = (
            np.asarray(
                np.gradient(
                    self.psf_flux,
                    np.median(np.diff(self.psf_row)).value / self.scale,
                    axis=(0, 1),
                )
            )
            * (self.sub_pixel_size / self.pixel_size).value
        )

    def _validate(self):
        """Method that finishes the set up of the PSF"""
        self.shape = self.psf_flux.shape[:2]
        self.ndims = len(self.dimension_names)
        # Set up 1D arrays
        if self.ndims == 1:
            setattr(
                self,
                self.dimension_names[0] + "1d",
                getattr(self, self.dimension_names[0]),
            )
            midpoint = getattr(self, self.dimension_names[0])
            midpoint = midpoint[len(midpoint) // 2]
            setattr(self, self.dimension_names[0] + "0d", midpoint)
        else:
            # Get 1D version of these grids
            dims = set(np.arange(self.ndims))
            for dim in np.arange(self.ndims):
                lp = getattr(self, self.dimension_names[dim]).transpose(
                    np.hstack([dim, list(dims - set([dim]))])
                )
                for d in range(self.ndims - 1):
                    lp = np.take(lp, 0, -1)
                s = np.argsort(lp.value)
                setattr(self, self.dimension_names[dim] + "1d", lp[s])
                # We have to do this to sort the axis, having them sorted will mean it's easier to interpolate later...
                reshape = np.hstack(
                    [np.hstack([dim, list(dims - set([dim]))]) + 2, 0, 1]
                )
                deshape = [
                    np.where(reshape == idx)[0][0] for idx in range(len(reshape))
                ]
                self.psf_flux = self.psf_flux.transpose(reshape)[s].transpose(deshape)

                midpoint = getattr(self, self.dimension_names[dim] + "1d")
                midpoint = midpoint[len(midpoint) // 2]
                setattr(self, self.dimension_names[dim] + "0d", midpoint)

        self.psf_column = (
            self.scale
            * (
                (np.arange(self.shape[1]) - self.shape[1] // 2)
                * u.pixel
                * self.sub_pixel_size
            )
            / self.pixel_size
        )
        self.psf_column -= np.median(np.diff(self.psf_column)) / 2
        self.psf_row = (
            self.scale
            * (
                (np.arange(self.shape[0]) - self.shape[0] // 2)
                * u.pixel
                * self.sub_pixel_size
            )
            / self.pixel_size
        )
        self.psf_row -= np.median(np.diff(self.psf_row)) / 2
        self.midpoint = tuple(
            [getattr(self, name + "0d") for name in self.dimension_names]
        )
        for dim, p in enumerate(self.dimension_names):
            lp = getattr(self, self.dimension_names[dim])
            setattr(self, p + "_bounds", [lp.min(), lp.max()])
        #        self._psf_flux_jitter = np.zeros_like(self._psf_flux)

    #        self._blur(self.blur_value)

    #       self.psf_flux = self._psf_flux_blur #+ self._psf_flux_jitter

    # def _blur(self, blur_value: Tuple):
    #     """Blurs the PSF using a Gaussian Kernel. Will update `self._psf_flux_blur` and `self._psf_flux_blur_grad`.

    #     Parameters:
    #     -----------
    #     blur_value: tuple of astropy quantities, must be in units of pixels
    #     """
    #     xstd, ystd = blur_value
    #     xstd, ystd = u.Quantity(xstd, "pixel"), u.Quantity(ystd, "pixel")

    #     xstd = ((self.pixel_size * xstd) / self.sub_pixel_size).value
    #     ystd = ((self.pixel_size * ystd) / self.sub_pixel_size).value
    #     a = deepcopy(self._psf_flux)
    #     if (xstd == 0) & (ystd == 0):
    #         print('christina')
    #         self._psf_flux_blur = deepcopy(self._psf_flux)
    #         self._psf_flux_blur_grad = (
    #             np.asarray(
    #                 np.gradient(
    #                     self._psf_flux_blur,
    #                     np.median(np.diff(self.psf_row)).value / self.scale,
    #                     axis=(0, 1),
    #                 )
    #             )
    #             * (self.sub_pixel_size / self.pixel_size).value
    #         )
    #         self.psf_flux = deepcopy(self._psf_flux)
    #         return
    #     s = a.shape
    #     a = a.reshape((s[0], s[1], np.prod(s[2:]).astype(int)))
    #     b = np.asarray(
    #         [
    #             convolve(
    #                 a[:, :, idx],
    #                 Gaussian2DKernel(
    #                     xstd,
    #                     ystd,
    #                 ),
    #             )
    #             for idx in range(a.shape[2])
    #         ]
    #     ).transpose([1, 2, 0])
    #     b = b.reshape(s)
    #     b /= b.sum(axis=(0, 1))[None, None]
    #     self._psf_flux_blur = b
    #     self._psf_flux_blur_grad = (
    #         np.asarray(
    #             np.gradient(
    #                 self._psf_flux_blur,
    #                 np.median(np.diff(self.psf_row)).value / self.scale,
    #                 axis=(0, 1),
    #             )
    #         )
    #         * (self.sub_pixel_size / self.pixel_size).value
    #     )
    #     self.psf_flux = self._psf_flux_blur  # + self._psf_flux_jitter
    #     return

    # def _jitter(self, row: npt.ArrayLike, column: npt.ArrayLike):
    #     """Jitters the PSF using gradients. Will update `self`

    #     Parameters:
    #     -----------
    #     row: npt.ArrayLike
    #         Row positions to jitter to. Will be downsampled
    #     column: npt.ArrayLike
    #         Column positions to jitter to. Will be downsampled
    #     """

    #     def grow(ar, ndims):
    #         for i in range(ndims):
    #             ar = ar[:, None]
    #         return ar

    #     def downsample(ar, npoints):
    #         nbin = np.ceil(ar.shape[0] / npoints).astype(int)
    #         a = np.asarray(
    #             [ar[idx * nbin : (idx + 1) * nbin] for idx in range(npoints)]
    #         )
    #         mean, weight = np.asarray([a1.mean() for a1 in a]), np.asarray(
    #             [len(a1) for a1 in a]
    #         ).astype(float)
    #         weight /= weight.max()
    #         return mean, weight

    #     if len(row) > 5:
    #         row, row_w = downsample(row, 5)
    #         column, column_w = downsample(column, 5)
    #     else:
    #         row_w = np.ones(len(row))
    #         column_w = np.ones(len(column))
    #     self._psf_flux_jitter *= 0
    #     if (row.sum() == 0) & (column.sum() == 0):
    #         self.psf_flux = self._psf_flux_blur + self._psf_flux_jitter
    #         return
    #     g1, g2 = self._psf_flux_blur_grad
    #     self._psf_flux_jitter = (
    #         g1 * grow(row * row_w, self.ndims + 2)
    #         + g2 * grow(column * column_w, self.ndims + 2)
    #     ).sum(axis=0)
    #     self.psf_flux = self._psf_flux_blur + self._psf_flux_jitter
    #     return

    def freeze_dimension(self, **kwargs):
        """Drop a dimension of the PSF model"""
        dnms = self.dimension_names.copy()
        duns = self.dimension_units.copy()
        PSF0 = self.psf_flux.copy()
        for key, point in kwargs.items():
            dim = np.where(np.in1d(dnms, key))[0][0]
            PSF0 = interpfunc(
                point.to(duns[dim]).value,
                getattr(self, dnms[dim] + "1d").value,
                reorder(PSF0, dim),
            )
            dnms.pop(dim)
            duns.pop(dim)

        X = {dnm: getattr(self, dnm) for dnm in dnms}
        dnms2 = self.dimension_names.copy()
        for key, point in kwargs.items():
            dim = np.where(np.in1d(dnms2, key))[0][0]
            for dnm in dnms:
                X[dnm] = X[dnm].transpose(
                    np.hstack([dim, list(set(np.arange(len(dnms2))) - set([dim]))])
                )[0]
            dnms2.pop(dim)
        psf2 = PSF(
            self.name,
            [X[dnm] for dnm in dnms],
            PSF0,
            dnms,
            duns,
            self.pixel_size,
            self.sub_pixel_size,
            freeze_dictionary=kwargs.copy(),
            #            blur_value=self.blur_value,
            extrapolate=self.extrapolate,
            scale=self.scale,
        )
        return psf2

    def integrate_spectrum(self, wavelength, spectrum, wavelength_grid=None):
        if wavelength_grid is None:
            wavelength_grid = self.trace_wavelength.to(wavelength.unit)
        dwavs = np.gradient(wavelength_grid.to(wavelength.unit))
        bounds0, bounds1 = (wavelength_grid.to(wavelength.unit) - dwavs / 2).value, (
            wavelength_grid.to(wavelength.unit) + dwavs / 2
        ).value
        sens = np.interp(
            wavelength,
            self.trace_wavelength.to(wavelength.unit),
            self.trace_sensitivity,
        )

        if (
            not ((spectrum.unit * sens.unit) * wavelength.unit).to(
                u.electron / u.second
            )
            == 1
        ):
            raise ValueError(
                "Spectrum must be in units equivalent to erg / (Angstrom s cm2)."
            )

        # Calculate the weighted flux contribution in each wavelength bin
        integrated_flux = np.ones(len(wavelength_grid))
        for idx in range(bounds0.shape[0]):
            x = np.linspace(bounds0[idx], bounds1[idx], 100)
            integrated_flux[idx] = np.trapz(
                np.hstack(
                    [0, *np.interp(x, wavelength.value, spectrum.value * sens.value), 0]
                ),
                np.hstack([x[0] - 1e-10, *x, x[-1] + 1e-10]),
            )
        integrated_flux *= u.electron / u.second

        return integrated_flux

    def integrate_wavelength(self, teff=5500, logg=4.5, **kwargs):
        if "wavelength" not in self.dimension_names:
            raise ValueError("PSF has no wavelength dimension.")

        # calculate spectrum of star
        w, s = get_phoenix_model(teff, logg=logg, jmag=9)
        integrated_flux = self.integrate_spectrum(
            w, s, wavelength_grid=self.wavelength1d
        )

        weighted_flux = (integrated_flux / integrated_flux.sum()).value

        dnms = self.dimension_names.copy()
        duns = self.dimension_units.copy()
        PSF0 = self.psf_flux.copy()
        dim = np.where(np.in1d(dnms, "wavelength"))[0][0]

        reshape = np.ones(PSF0.ndim, dtype=int)
        reshape[dim + 2] = weighted_flux.shape[0]
        weighted_flux = weighted_flux.reshape(tuple(reshape.astype(int)))
        PSF0 = (PSF0 * weighted_flux).sum(axis=dim + 2)

        X = {
            dnm: getattr(self, dnm)
            for dnm in list(set(list(dnms)) - set(["wavelength"]))
        }

        for dnm in list(set(list(dnms)) - set(["wavelength"])):
            X[dnm] = X[dnm].transpose(
                np.hstack([dim, list(set(np.arange(len(dnms))) - set([dim]))])
            )[0]

        dnms.pop(dim)
        duns.pop(dim)

        psf2 = PSF(
            self.name,
            [X[dnm] for dnm in dnms],
            PSF0,
            dnms,
            duns,
            self.pixel_size,
            self.sub_pixel_size,
            freeze_dictionary=kwargs.copy(),
            #            blur_value=self.blur_value,
            extrapolate=self.extrapolate,
            scale=self.scale,
        )
        return psf2

    def __repr__(self):
        freeze_dictionary = (
            f" (Frozen: {', '.join([f'{key}: {item:.3f}' for key, item in self.freeze_dictionary.items()])})"
            if len(self.freeze_dictionary) != 0
            else ""
        )
        return f"{self.ndims}D PSF Model [{', '.join(self.dimension_names)}]{freeze_dictionary}"

    @staticmethod
    def from_name(
        name: str,
        transpose: bool = False,
        scale: int = 1,
        bin: int = 1,
        #        blur_value: Tuple = (0 * u.pixel, 0 * u.pixel),
    ):
        """Open a PSF file based on the detector name"""
        if name.lower() in ["gauss", "gaussian", "test"]:
            p = PSF.from_file(
                "gaussian",
                f"{PACKAGEDIR}/data/pandora_gaussian_psf.fits",
                transpose=transpose,
                extrapolate=True,
                scale=scale,
                bin=bin,
            )
            return p

        if name.lower() in ["vis", "visda", "visible"]:
            p = PSF.from_file(
                "visda",
                f"{PACKAGEDIR}/data/pandora_vis_2024-05.fits",
                transpose=transpose,
                extrapolate=True,
                scale=scale,
                bin=bin,
            ).integrate_wavelength(teff=5500, logg=4.5)
            # p = p.freeze_dimension(wavelength=p.wavelength0d)
            #            p.blur_value = blur_value
            #            p._blur(blur_value=p.blur_value)
            return p

        elif name.lower() in ["nir", "nirda", "ir"]:
            p = PSF.from_file(
                "nirda",
                f"{PACKAGEDIR}/data/pandora_nir_2024-05.fits",
                transpose=transpose,
                extrapolate=True,
                scale=scale,
                bin=bin,
            )
            # p = p.freeze_dimension(temperature=-5 * u.deg_C)
            # if we were modeling full focal plane we'd change this
            # p = p.freeze_dimension(row=p.row0d, column=p.column0d)
            #            p.blur_value = blur_value
            #            p._blur(blur_value=p.blur_value)
            return p
        else:
            raise ValueError(f"No such PSF as `{name}`")

    def _add_trace_params(self):
        if self.name == "visda":
            detector = ps.VisibleDetector()
        elif self.name == "nirda":
            detector = ps.NIRDetector()
        else:
            raise ValueError(f"Can not parse name {self.name}.")
        self._trace_sensitivity, self._trace_pixel, self._trace_wavelength = (
            detector.trace_sensitivity,
            detector.trace_pixel,
            detector.trace_wavelength,
        )

    @property
    def trace_sensitivity(self):
        if not hasattr(self, "_trace_sensitivity"):
            self._add_trace_params()
        return self._trace_sensitivity

    @property
    def trace_wavelength(self):
        if not hasattr(self, "_trace_wavelength"):
            self._add_trace_params()
        return self._trace_wavelength

    @property
    def trace_pixel(self):
        if not hasattr(self, "_trace_pixel"):
            self._add_trace_params()
        return self._trace_pixel

    @staticmethod
    def from_file(
        name: str,
        filename: str = f"{PACKAGEDIR}/data/pandora_vis_2024-05.fits",
        transpose: bool = False,
        #        blur_value: Tuple = (0 * u.pixel, 0 * u.pixel),
        extrapolate: bool = False,
        scale: int = 1,
        bin: int = 1,
    ):
        """Build a PSF class from an input fits file.

        Parameters:
        -----------
        filename: str
            Filename of PSF fits file. PSF cube must have a shape such that the first
            two dimensions are the x and y extent of the PSF. Defaults to visible PSF.
        transpose: bool
            Transpose the input file, i.e. rotate 90 degrees
        blur_value: tuple:
            A tuple describing the amount of motion blur in each axis. Optional, defaults to zero.

        Returns:
        --------
        psf : pandorapsf.PSF
            Returns a PSF object
        """
        hdu = fits.open(filename)
        pixel_size = hdu[0].header["PIXSIZE"] * u.micron / u.pix
        sub_pixel_size = hdu[0].header["SUBPIXSZ"] * u.micron / u.pix

        # LLNL's matlab files are COLUMN-major
        # This should make the array ROW-major
        replace = {"x": "column", "y": "row"}
        dimension_names = [
            replace[i.name.lower()] if i.name.lower() in replace else i.name.lower()
            for i in hdu[2:]
        ]

        if "row" in dimension_names:
            l = (
                np.where(np.asarray(dimension_names) == "row")[0][0],
                np.where(np.asarray(dimension_names) == "column")[0][0],
            )
            l = np.hstack(
                [l, list(set(list(np.arange(len(hdu) - 2))) - set(l))]
            ).astype(int)
        else:
            l = np.arange(len(hdu) - 2).astype(int)

        psf_flux = hdu[1].data.transpose(np.hstack([1, 0, *l + 2]))
        dimension_names = [dimension_names[l1] for l1 in l]
        dimension_units = [u.Unit(hdu[l1].header["UNIT"]) for l1 in l + 2]
        X = [hdu[l1].data.transpose(l) * u.Unit(hdu[l1].header["UNIT"]) for l1 in l + 2]

        return PSF(
            name,
            X,
            psf_flux,
            dimension_names,
            dimension_units,
            pixel_size,
            sub_pixel_size,
            transpose=transpose,
            #            blur_value=blur_value,
            extrapolate=extrapolate,
            scale=scale,
            bin=bin,
        )

    def _get_dim(self, dim: Union[int, str], dimension_names=None):
        """Return the numeric dimension of an input int or string"""
        if isinstance(dim, int):
            if (dim > self.ndims) | (dim < 0):
                raise ValueError(f"No dimension `{dim}`")
            l = dim
        elif isinstance(dim, str):
            if dimension_names is None:
                dimension_names = self.dimension_names
            l = np.where(np.asarray(dimension_names) == dim.lower())[0]
            if len(l) == 0:
                raise ValueError(f"No dimension `{dim}`")
            l = l[0]
        return l

    def _check_bounds(self, **kwargs):
        """Check a given point has the right shape, units and values in bounds"""
        cleaned = {}
        for key, value in kwargs.items():
            l = np.in1d(self.dimension_names, key)
            if not l.any():
                raise KeyError(f"`{key}` not in PSF dimensions.")
            dim = np.where(l)[0][0]
            value = u.Quantity(value, self.dimension_units[dim])
            bounds = getattr(self, self.dimension_names[dim] + "_bounds")
            if (value.value < bounds[0].value) | (value.value > bounds[1].value):
                if not self.extrapolate:
                    raise OutOfBoundsError(
                        f"Point ({value}) out of {self.dimension_names[dim]} bounds."
                    )
                else:
                    if value.value < bounds[0].value:
                        cleaned[key] = bounds[0]
                    elif value.value > bounds[1].value:
                        cleaned[key] = bounds[1]
            else:
                cleaned[key] = value
        return cleaned

    def psf(self, gradients=False, **kwargs):
        """Interpolate the PSF cube to a particular point

        Parameters
        ----------
        args: dict
            Dictionary of arguments to pass, set each dimension name to a value
        Returns
        -------
        ar : np.ndarray of shape self.shape
            The interpolated PSF
        """
        if self.check_bounds:
            kwargs = self._check_bounds(**kwargs)
        d = kwargs.copy()
        PSF0 = self.psf_flux
        if gradients:
            dPSF0 = self.psf_flux_grad[0]
            dPSF1 = self.psf_flux_grad[1]
        for key in self.dimension_names:
            value = d.pop(key, getattr(self, key + "0d"))
            PSF0 = interpfunc(
                value.value,
                getattr(self, key + "1d").value,
                PSF0,
            )
            if gradients:
                dPSF0 = interpfunc(
                    value.value,
                    getattr(self, key + "1d").value,
                    dPSF0,
                )
                dPSF1 = interpfunc(
                    value.value,
                    getattr(self, key + "1d").value,
                    dPSF1,
                )
        integral = PSF0.sum()
        if gradients:
            return (
                PSF0 / integral,
                (dPSF0 - dPSF0.mean()) / integral,
                (dPSF1 - dPSF1.mean()) / integral,
            )
        return PSF0 / integral

    def prf(self, row, column, gradients=False, **kwargs):
        """
        Bins the PSF down to the pixel scale.

        Parameters
        ----------
        args: dict
            Dictionary of arguments to pass, set each dimension name to a value

        Returns
        -------
        row: np.ndarray
            Array of integer row positions
        column: np.ndarray
            Array of integer column positions
        psf: np.ndarray
            2D array of PRF flux values with shape (nrows, ncolumns)
        """
        if self.check_bounds:
            test1 = np.in1d(list(kwargs.keys()), self.dimension_names)
            if not test1.all():
                raise ValueError(
                    f"Pass only dimension names from {self.dimension_names}"
                )
        row, column = u.Quantity(row, u.pixel), u.Quantity(column, u.pixel)
        if "row" in self.dimension_names:
            if gradients:
                psf0, dpsf0, dpsf1 = self.psf(
                    row=row, column=column, gradients=True, **kwargs
                )
            else:
                psf0 = self.psf(row=row, column=column, **kwargs)
        else:
            if gradients:
                psf0, dpsf0, dpsf1 = self.psf(gradients=True, **kwargs)
            else:
                psf0 = self.psf(**kwargs)
        rb, cb, psfb = bin_prf(
            psf0,
            self.psf_row.value,
            self.psf_column.value,
            (row.value, column.value),
            normalize=False,
        )
        integral = np.sum(psfb)
        if gradients:
            _, _, dpsf0b = bin_prf(
                dpsf0,
                self.psf_row.value,
                self.psf_column.value,
                (row.value, column.value),
                normalize=False,
            )
            _, _, dpsf1b = bin_prf(
                dpsf1,
                self.psf_row.value,
                self.psf_column.value,
                (row.value, column.value),
                normalize=False,
            )
            # return rb, cb, psfb/integral, (dpsf0b - dpsf0b.mean())/integral, (dpsf1b - dpsf1b.mean())/integral
            # dpsf0, dpsf1 = np.gradient(psfb)
            dpsf0b -= dpsf0b.mean()
            dpsf1b -= dpsf1b.mean()
            return rb, cb, psfb / integral, dpsf0b / integral, dpsf1b / integral
        return rb, cb, psfb / integral

    # def _bin_prf(self, psf0, row, column, normalize=True):
    #     mod = (self.psf_column.value + column.value) % 1
    #     cyc = ((self.psf_column.value + column.value) - mod).astype(int)
    #     colbin = np.unique(cyc)
    #     psf1 = np.asarray(
    #         [psf0[:, cyc == c].sum(axis=1) / (cyc == c).sum() for c in colbin]
    #     ).T
    #     mod = (self.psf_row.value + row.value) % 1
    #     cyc = ((self.psf_row.value + row.value) - mod).astype(int)
    #     rowbin = np.unique(cyc)
    #     psf2 = np.asarray(
    #         [psf1[cyc == c].sum(axis=0) / (cyc == c).sum() for c in rowbin]
    #     )
    #     if normalize:
    #         psf2 /= psf2.sum()
    #     return rowbin.astype(int), colbin.astype(int), psf2


def interpfunc(l, lp, PSF0):
    if l in lp:
        PSF1 = PSF0[:, :, np.where(lp == l)[0][0]]
    elif l < lp[0]:
        PSF1 = PSF0[:, :, 0]
    elif l > lp[-1]:
        PSF1 = PSF0[:, :, -1]
    else:
        # Find the two closest frames
        d = np.argsort(np.abs(lp - l))[:2]
        d = d[np.argsort(lp[d])]
        # Linearly interpolate
        slope = (PSF0[:, :, d[0]] - PSF0[:, :, d[1]]) / (lp[d[0]] - lp[d[1]])
        PSF1 = PSF0[:, :, d[1]] + (slope * (l - lp[d[1]]))
    return PSF1


def reorder(ar: np.ndarray, dim: int = 0):
    """Reorders a PSF array so that a different dimension is in the front, this helps when we interpolate."""
    if not isinstance(dim, (int, list, np.int_)):
        raise ValueError("Pass an `int` or a `list` of `int`s.")
    if not (ar.ndim - 2) > (np.max(dim)):
        raise ValueError(
            f"No dimension {[d + 2 for d in dim]} in array shape {ar.shape}"
        )
    if np.min(dim) < 0:
        raise ValueError("Can not reorder the first two dimensions.")
    if not hasattr(dim, "__iter__"):
        dim = [dim]
    cdim = [d + 2 for d in dim]
    l = set(np.arange(ar.ndim)[2:]) - set(cdim)
    return ar.transpose(np.hstack([0, 1, cdim, list(l)]).astype(int))


class OutOfBoundsError(Exception):
    """Exception raised if a point is out of bounds for this PSF"""

    pass
